from functools import wraps

from pyrio.exceptions import IllegalStateError

TERMINAL_FUNCTIONS = [
    "for_each",
    "reduce",
    "count",
    "min",
    "max",
    "sum",
    "average",
    "find_first",
    "find_any",
    "take_first",
    "take_last",
    "take_nth",
    "any_match",
    "all_match",
    "none_match",
    "compare_with",
    "all_equal",
    "quantify",
    "group_by",
    "collect",
    "to_list",
    "to_tuple",
    "to_set",
    "to_dict",
    "to_string",
    "save",
]


def pre_call(function_decorator):
    """Applies a function decorator to all callable methods"""

    def decorator(cls):
        for name, obj in vars(cls).items():
            if callable(obj):
                setattr(cls, name, function_decorator(obj))
        return cls

    return decorator


def handle_consumed(func):
    """Prevents operations on consumed streams and auto-closes after terminal operations"""

    @wraps(func)
    def wrapper(*args, **kw):
        from pyrio.streams.base_stream import BaseStream

        stream = args[0] if args else None
        if not (stream and isinstance(stream, BaseStream)):
            return func(*args, **kw)  # pragma: no cover

        is_consumed = getattr(stream, "_is_consumed", None)
        if is_consumed and func.__name__ != "close":
            raise IllegalStateError("Stream object already consumed")

        result = func(*args, **kw)
        if not is_consumed and func.__name__ in TERMINAL_FUNCTIONS:
            stream.close()
        return result

    return wrapper



from pyrio.exceptions import NoSuchElementError, NoneTypeError


class Optional:
    """Container object which may (or may not) contain a non-null value"""

    def __init__(self, element):
        self._element = element

    def __str__(self):
        return f"Optional[{self._element}]"

    @staticmethod
    def empty():
        """Creates empty Optional"""
        return Optional(None)

    @staticmethod
    def of(element):
        """Creates Optional describing given non-null value"""
        if element is None:
            raise NoneTypeError("Value cannot be None")
        return Optional(element)

    @staticmethod
    def of_nullable(element):
        """
        Returns an Optional describing the given value, if non-null,
        otherwise returns an empty Optional
        """
        return Optional(element)

    def get(self):
        """If a value is present, returns the value, otherwise raises an Exception"""
        if self.is_empty():
            raise NoSuchElementError("Optional is empty")
        return self._element

    def is_present(self):
        """Returns bool whether a value is present"""
        return not self.is_empty()

    def is_empty(self):
        """Returns bool whether the Optional is empty"""
        return self._element is None

    def if_present(self, action):
        """Performs given action with the value if the Optional is not empty"""
        if self.is_present():
            action(self.get())

    def if_present_or_else(self, action, empty_action):
        """Performs given action with the value if the Optional is not empty,
        otherwise calls fallback 'empty_action'
        """
        if self.is_present():
            action(self.get())
        else:
            empty_action()

    def or_else(self, value):
        """
        Returns the value if present, or a provided argument otherwise.
        Safe alternative to get() method
        """
        return self._element if self.is_present() else value

    def or_else_get(self, supplier):
        """
        Returns the value if present, or calls a 'supplier' function otherwise.
        Safe alternative to get() method
        """
        return self._element if self.is_present() else supplier()

    def or_else_raise(self, supplier=None):
        """
        Returns the value if present,
        otherwise throws an exception produced by the exception supplying function
        (if such is provided by the user) or NoSuchElementError
        """
        if supplier is None:

            def supplier():
                raise NoSuchElementError("Optional is empty")

        return self._element if self.is_present() else supplier()



from collections.abc import Mapping

from pyrio.iterators import StreamGenerator
from pyrio.decorators import handle_consumed, pre_call
from pyrio.utils import DictItem, Optional
from pyrio.exceptions import IllegalStateError, UnsupportedTypeError, NoneTypeError


@pre_call(handle_consumed)
class BaseStream:
    """Base class for Stream objects; describes core supported operations"""

    def __init__(self, iterable):
        if iterable is None:
            raise NoneTypeError("Cannot create Stream from None")
        self._iterable = iterable
        self._is_consumed = False
        self._on_close_handler = None

    def __iter__(self):
        return iter(self.iterable)

    @property
    def iterable(self):
        if isinstance(self._iterable, Mapping):
            self._iterable = tuple(DictItem(k, v) for k, v in self._iterable.items())
        return self._iterable

    @iterable.setter
    def iterable(self, value):
        self._iterable = value

    def concat(self, *streams):
        """Concatenates several streams together or adds new streams/collections to the current one"""
        self.iterable = StreamGenerator.concat(self.iterable, *streams)
        return self

    def prepend(self, iterable):
        """Prepends iterable to current stream"""
        self.iterable = StreamGenerator.concat(iterable, self.iterable)
        return self

    def filter(self, predicate):
        """Filters values in stream based on given predicate function"""
        self.iterable = StreamGenerator.filter(self.iterable, predicate)
        return self

    def map(self, mapper):
        """Returns a stream consisting of the results of applying the given function to the elements of this stream"""
        self.iterable = StreamGenerator.map(self.iterable, mapper)
        return self

    def filter_map(self, mapper, *, discard_falsy=False):
        """Filters out all None or falsy values and applies mapper function to the elements of the stream"""
        self.iterable = StreamGenerator.filter_map(self.iterable, mapper, discard_falsy)
        return self

    def flat_map(self, mapper):
        """Maps each element of the stream and yields the elements of the produced iterators"""
        self.iterable = StreamGenerator.flat_map(self.iterable, mapper)
        return self

    def flatten(self):
        """Converts a Stream of multidimensional collection into a one-dimensional"""
        self.iterable = StreamGenerator.flatten(self.iterable)
        return self

    def peek(self, operation):
        """Performs the provided operation on each element of the stream without consuming it"""
        self.iterable = StreamGenerator.peek(self.iterable, operation)
        return self

    def distinct(self):
        """Returns a stream with the distinct elements of the current one"""
        self.iterable = StreamGenerator.distinct(self.iterable)
        return self

    def count(self):
        """Returns the count of elements in the stream"""
        return len(tuple(self.iterable))

    def _validate_numeric_data(self, op):
        data = tuple(self.iterable)
        if not all(isinstance(x, (int, float)) or x is None for x in data):
            raise ValueError(f"Cannot apply {op} on non-number elements")
        return [x for x in data if x is not None]

    def sum(self):
        """Sums the elements of the stream"""
        valid_data = self._validate_numeric_data(self.sum.__name__)
        return sum(valid_data) if valid_data else 0

    def average(self):
        """Returns the average value of elements in the stream"""
        valid_data = self._validate_numeric_data(self.average.__name__)
        return sum(valid_data) / len(valid_data) if valid_data else 0

    def skip(self, count):
        """Discards the first n elements of the stream and returns a new stream with the remaining ones"""
        if count < 0:
            raise ValueError("Skip count cannot be negative")
        self.iterable = StreamGenerator.skip(self.iterable, count)
        return self

    def limit(self, count):
        """Returns a stream with the first n elements, or fewer if the underlying iterator ends sooner"""
        if count < 0:
            raise ValueError("Limit count cannot be negative")
        self.iterable = StreamGenerator.limit(self.iterable, count)
        return self

    def head(self, count):
        """Alias for 'limit'"""
        if count < 0:
            raise ValueError("Head count cannot be negative")
        self.iterable = StreamGenerator.limit(self.iterable, count)
        return self

    def tail(self, count):
        """Returns a stream with the last n elements, or fewer if the underlying iterator ends sooner"""
        if count < 0:
            raise ValueError("Tail count cannot be negative")
        self.iterable = StreamGenerator.tail(self.iterable, count)
        return self

    def take_while(self, predicate):
        """Returns a stream that yields elements based on a predicate"""
        self.iterable = StreamGenerator.take_while(self.iterable, predicate)
        return self

    def drop_while(self, predicate):
        """Returns a stream that skips elements based on a predicate and yields the remaining ones"""
        self.iterable = StreamGenerator.drop_while(self.iterable, predicate)
        return self

    def take_first(self, default=None):
        """Returns Optional with the first element of the stream or a default value"""
        return Optional.of_nullable(next(iter(self.iterable), default))

    def take_last(self, default=None):
        """Returns Optional with the last element of the stream or a default value"""
        from collections import deque

        last_item = deque(self.iterable, maxlen=1)
        return Optional.of_nullable(last_item[0] if last_item else default)

    def sort(self, comparator=None, *, reverse=False):
        """
        Sorts the elements of the current stream according to natural order or based on the given comparator.
        If 'reverse' flag is True, the elements are sorted in descending order
        """
        self.iterable = StreamGenerator.sort(self.iterable, comparator, reverse)
        return self

    def reverse(self, comparator=None):
        """
        Sorts the elements of the current stream in descending order.
        Alias for 'sort(comparator, reverse=True)'
        """
        self.iterable = StreamGenerator.sort(self.iterable, comparator, reverse=True)
        return self

    def find_first(self, predicate=None):
        """
        Searches for an element of the stream that satisfies a predicate.
        Returns an Optional with the first found value, if any, or None
        """
        return Optional.of_nullable(next(filter(predicate, self.iterable), None))

    def find_any(self, predicate=None):
        """
        Searches for an element of the stream that satisfies a predicate.
        Returns an Optional with some of the found values, if any, or None
        """
        import random

        if predicate:
            self.filter(predicate)
        try:
            return Optional.of(random.choice(list(self.iterable)))
        except IndexError:
            return Optional.of_nullable(None)

    def any_match(self, predicate):
        """Returns whether any elements of the stream match the given predicate"""
        return any(predicate(i) for i in self.iterable)

    def all_match(self, predicate):
        """Returns whether all elements of the stream match the given predicate"""
        return all(predicate(i) for i in self.iterable)

    def none_match(self, predicate):
        """Returns whether no elements of the stream match the given predicate"""
        return not any(predicate(i) for i in self.iterable)

    def min(self, comparator=None, default=None):
        """Returns the minimum element of the stream according to the given comparator"""
        return Optional.of_nullable(min(self.iterable, key=comparator, default=default))

    def max(self, comparator=None, default=None):
        """Returns the maximum element of the stream according to the given comparator"""
        return Optional.of_nullable(max(self.iterable, key=comparator, default=default))

    def for_each(self, operation):
        """Performs an action for each element of this stream"""
        for i in self.iterable:
            operation(i)

    def enumerate(self, start=0):
        """
        Returns each element of the Stream preceded by his corresponding index
        (by default starting from 0 if not specified otherwise)
        """
        self.iterable = StreamGenerator.enumerate(self.iterable, start)
        return self

    def reduce(self, accumulator, identity=None):
        """
        Reduces the elements to a single one, by repeatedly applying a reducing operation.
        Returns Optional with the result, if any, or None
        """
        curr_iter = iter(self.iterable)
        try:
            if identity is None:
                identity = next(curr_iter)
        except StopIteration:
            return Optional.of_nullable(identity)

        for i in curr_iter:
            identity = accumulator(identity, i)
        return Optional.of_nullable(identity)

    def compare_with(self, other, comparator=None):
        """Compares current stream with another one based on a given comparator"""
        return not any(
            (comparator and not comparator(i, j)) or i != j for i, j in zip(self.iterable, other)
        )

    # ### collectors ###
    def collect(self, collection_type, dict_collector=None, dict_merger=None, str_delimiter=", "):
        """
        Returns a collection from the stream.

        In case of dict:
        The 'dict_collector' function receives an element from the stream and returns a (key, value) pair or a DictItem
        specifying how the dict should be constructed.

        The 'dict_merger' functions indicates in the case of a collision (duplicate keys), which entry should be kept.
        E.g. lambda old, new: new

        In case of str:
        Concatenates the elements of the Stream, separated by the specified 'str_delimiter'
        """
        import builtins

        match collection_type:
            case builtins.tuple:
                return self.to_tuple()
            case builtins.list:
                return self.to_list()
            case builtins.set:
                return self.to_set()
            case builtins.dict:
                return self.to_dict(dict_collector, dict_merger)
            case builtins.str:
                return self.to_string(str_delimiter)
            case _:
                raise ValueError("Invalid collection type")

    def to_list(self):
        """Returns a list of the elements of the current stream"""
        return list(self.iterable)

    def to_tuple(self):
        """Returns a tuple of the elements of the current stream"""
        return tuple(self.iterable)

    def to_set(self):
        """Returns a set of the elements of the current stream"""
        return set(self.iterable)

    def to_dict(self, collector=None, merger=None):
        """
        Returns a dict of the elements of the current stream.

        The 'collector' function receives an element from the stream and returns a (key, value) pair or a DictItem
        specifying how the dict should be constructed.

        The 'merger' functions indicates in the case of a collision (duplicate keys), which entry should be kept.
        E.g. lambda old, new: new
        """
        result = {}
        source = (collector(i) for i in self.iterable) if collector else self.iterable
        for item in source:
            k, v = self._unpack_dict_item(item)
            if k in result:
                if merger is None:
                    raise IllegalStateError(f"Key '{k}' already exists")
                v = merger(result[k], v)
            result[k] = v
        return result

    def _unpack_dict_item(self, item):  # noqa
        match item:
            case tuple():
                return item[0], item[1]
            case DictItem():
                # let's not make unnecessary calls to property getters
                return item._key, item._value  # noqa
            case _:
                raise UnsupportedTypeError(
                    f"Cannot create dict items from '{item.__class__.__name__}' type"
                )

    def to_string(self, delimiter=", "):
        """Concatenates the elements of the Stream, separated by the specified delimiter"""
        return self._join(delimiter)

    def group_by(self, classifier=None, collector=None):
        """
        Performs a "group by" operation on the elements of the stream according to a classification function.
        Returns the results in a dict built using collector function
        (optionally provided by the user or via a default one)
        """
        if collector is None:
            return {key: list(group) for key, group in self._group_by(classifier)}

        result = {}
        for key, group in self._group_by(classifier):
            key, group = collector(key, list(group))
            if hasattr(group, "__iter__"):
                if key not in result:
                    result[key] = []
                result[key] += group
            else:
                result[key] = group
        return result

    def _group_by(self, classifier=None):
        # https://docs.python.org/3/library/itertools.html#itertools.groupby
        classifier = (lambda x: x) if classifier is None else classifier
        iterator = iter(self.iterable)
        exhausted = False

        def _grouper(target_key):  # noqa
            nonlocal curr_value, curr_key, exhausted
            yield curr_value
            for curr_value in iterator:
                curr_key = classifier(curr_value)
                if curr_key != target_key:
                    return
                yield curr_value
            exhausted = True

        try:
            curr_value = next(iterator)
        except StopIteration:
            return
        curr_key = classifier(curr_value)

        while not exhausted:
            target_key = curr_key
            curr_group = _grouper(target_key)
            yield curr_key, curr_group
            if curr_key == target_key:
                for _ in curr_group:
                    pass

    def quantify(self, predicate=bool):
        """Count how many of the elements are Truthy or evaluate to True based on a given predicate"""
        return sum(self.map(predicate))

    def close(self):
        """Closes the stream, causing the provided close handler to be called"""
        if self._on_close_handler:
            self._on_close_handler()
        self._is_consumed = True

    def on_close(self, handler):
        """Returns an equivalent stream with an additional close handler"""
        self._on_close_handler = handler
        return self

    # ### let's look nice ###
    def __repr__(self):
        return f"{self.__class__.__name__}.of({self._join()})"

    def _join(self, delimiter=", "):
        return delimiter.join(str(i) for i in self.iterable)



from pyrio.decorators import pre_call, handle_consumed
from pyrio.streams import BaseStream
from pyrio.iterators import StreamGenerator, ItertoolsMixin


@pre_call(handle_consumed)
class Stream(BaseStream, ItertoolsMixin):
    """Abstraction over a sequence of elements supporting sequential aggregate operations"""

    @classmethod
    def of(cls, *iterable):
        """Creates Stream from args"""
        return cls(iterable)

    @classmethod
    def of_nullable(cls, iterable):
        """Creates Stream from args if iterable is not None; otherwise returns empty Stream"""
        if iterable is None:
            return cls.empty()
        return cls(iterable)

    @classmethod
    def empty(cls):
        """Creates empty Stream"""
        return cls([])

    @classmethod
    def iterate(cls, seed, operation, condition=None):
        """Creates infinite ordered Stream"""
        return cls(StreamGenerator.iterate(seed, operation, condition))

    @classmethod
    def generate(cls, supplier):
        """Creates infinite unordered Stream with values generated by given supplier function"""
        return cls(StreamGenerator.generate(supplier))

    @classmethod
    def constant(cls, element):
        """Creates infinite Stream with given value"""
        return cls.generate(lambda: element)

    @classmethod
    def from_range(cls, start, stop, step=1):
        """Creates Stream from start (inclusive) to stop (exclusive) by an incremental step"""
        return cls(StreamGenerator.range(start, stop, step))

    # NB: handle_consumed decorator needs access to toggle flag
    def take_nth(self, idx, default=None):
        """Returns Optional with the nth element of the stream or a default value"""
        return super().take_nth(idx, default)

    def all_equal(self, key=None):
        """Returns True if all elements of the stream are equal to each other"""
        return super().all_equal(key)



from collections.abc import Iterable

from pyrio.decorators import map_dict_items


class StreamGenerator:
    """Helper class wrapping generator-based operations for lazy evaluation"""

    @staticmethod
    @map_dict_items
    def concat(*streams):
        """Concatenates multiple iterables into a single sequence"""
        for iterable in streams:
            yield from iterable

    @staticmethod
    def filter(iterable, predicate):
        """Yields elements that satisfy the predicate"""
        for i in iterable:
            if predicate(i):
                yield i

    @staticmethod
    def map(iterable, mapper):
        """Applies mapper function to each element"""
        for i in iterable:
            yield mapper(i)

    @staticmethod
    def filter_map(iterable, mapper, discard_falsy=False):
        """Filters out None (or falsy) values and applies mapper to remaining elements"""
        for i in iterable:
            if (not discard_falsy and i is not None) or (discard_falsy and i):
                yield mapper(i)

    @staticmethod
    def flat_map(iterable, mapper):
        """Applies mapper and flattens the resulting iterables"""
        for i in iterable:
            yield from mapper(i)

    @classmethod
    def flatten(cls, iterable):
        """Recursively flattens nested iterables into a single sequence"""
        for i in iterable:
            if isinstance(i, str) or not isinstance(i, Iterable):
                yield i
            else:
                yield from cls.flatten(i)

    @staticmethod
    def peek(iterable, operation):
        """Performs operation on each element without consuming the stream"""
        for i in iterable:
            operation(i)
            yield i

    @staticmethod
    def iterate(seed, operation, condition=None):
        """Generates sequence by repeatedly applying operation to seed"""
        if condition is None:
            condition = lambda _: True  # noqa
        while condition(seed):
            yield seed
            seed = operation(seed)

    @staticmethod
    def generate(supplier):
        """Generates infinite sequence using supplier function"""
        while True:
            yield supplier()

    @staticmethod
    def range(start, stop, step=1):
        """Yields values from start to stop with given step"""
        for i in range(start, stop, step):
            yield i

    @staticmethod
    def distinct(iterable):
        """Yields unique elements preserving first occurrence order"""
        elements = set()
        for i in iterable:
            if i not in elements:
                elements.add(i)
                yield i

    @staticmethod
    def skip(iterable, count):
        """Skips first n elements and yields the rest"""
        for i in iterable:
            if count > 0:
                count -= 1
            else:
                yield i

    @staticmethod
    def limit(iterable, count):
        """Yields at most n elements"""
        for i in iterable:
            if count == 0:
                break
            yield i
            count -= 1

    @staticmethod
    def tail(iterable, count):
        """Yields the last n elements"""
        import collections

        for i in collections.deque(iterable, maxlen=count):
            yield i

    @staticmethod
    def take_while(iterable, predicate):
        """Yields elements while predicate is true"""
        for i in iterable:
            if not predicate(i):
                break
            yield i

    @staticmethod
    def drop_while(iterable, predicate):
        """Skips elements while predicate is true, then yields the rest"""
        iterator = iter(iterable)
        for x in iterator:
            if not predicate(x):
                yield x
                break

        for x in iterator:
            yield x

    @staticmethod
    def sort(iterable, comparator=None, reverse=False):
        """Yields elements in sorted order"""
        for i in sorted(iterable, key=comparator, reverse=reverse):
            yield i

    @staticmethod
    def enumerate(iterable, start=0):
        """Yields index-element pairs starting from given index"""
        for i, item in enumerate(iterable, start):
            yield i, item



import importlib
import shutil
from contextlib import contextmanager
from pathlib import Path

from aldict import AliasDict

from pyrio.utils import DictItem
from pyrio.streams import BaseStream, Stream
from pyrio.exceptions import NoneTypeError

TEMP_PATH = "{file_path}.tmp"


DSV_CONFIG = {
    ".csv": {
        "delimiter": ",",
    },
    ".tsv": {
        "delimiter": "\t",
    },
}

MAPPING_READ_CONFIG = AliasDict(
    {
        ".toml": {
            "import_mod": "tomllib",
            "callable": "load",
            "read_mode": "rb",
        },
        ".json": {
            "import_mod": "json",
            "callable": "load",
            "read_mode": "r",
        },
        ".yaml": {
            "import_mod": "yaml",
            "callable": "safe_load",
            "read_mode": "r",
        },
        ".xml": {
            "import_mod": "xmltodict",
            "callable": "parse",
            "read_mode": "rb",
        },
    },
    aliases={".yaml": ".yml"},
)

MAPPING_WRITE_CONFIG = AliasDict(
    {
        ".toml": {
            "import_mod": "tomli_w",
            "callable": "dump",
            "write_mode": "wb",
            "default_null_handler": lambda x: DictItem(x.key, "N/A") if x.value is None else x,
        },
        ".json": {
            "import_mod": "json",
            "callable": "dump",
            "write_mode": "w",
            "default_null_handler": None,
        },
        ".yaml": {
            "import_mod": "yaml",
            "callable": "dump",
            "write_mode": "w",
            "default_null_handler": None,
        },
        ".xml": {
            "import_mod": "xmltodict",
            "callable": "unparse",
            "write_mode": "w",
            "default_null_handler": None,
        },
    },
    aliases={".yaml": ".yml"},
)


class FileStream(BaseStream):
    """Derived Stream class for querying files; maps file content to im-memory dict structures and vice versa"""

    # Dirty deeds for a nice-looking API
    def __init__(self, file_path):  # noqa
        """Creates Stream from a file"""
        pass

    def __new__(cls, file_path, f_open_options=None, f_read_options=None, **kwargs):
        obj = super().__new__(cls)
        if file_path is None:
            raise NoneTypeError("File path cannot be None")
        file_handler = None
        try:
            file_handler, iterable = cls._read_file(
                file_path, f_open_options, f_read_options, **kwargs
            )
            super(cls, obj).__init__(iterable)
            obj._file_path = file_path
            obj._file_handler = file_handler
            obj._on_close_handler = (
                lambda: obj._file_handler.close() if not obj._file_handler.closed else None
            )
            return obj
        except Exception:
            if file_handler is not None and not file_handler.closed:
                file_handler.close()
            raise

    @classmethod
    def process(cls, file_path, *, f_open_options=None, f_read_options=None, **kwargs):
        """Creates Stream from a file with advanced 'reading' options passed by the user"""
        return cls.__new__(cls, file_path, f_open_options, f_read_options, **kwargs)

    # ### reading from file ###
    @classmethod
    def _read_file(cls, file_path, f_open_options=None, f_read_options=None, **kwargs):
        path = cls._get_file_path(file_path)

        f_open_options = f_open_options or {}
        f_read_options = f_read_options or {}

        if (suffix := path.suffix) in DSV_CONFIG:
            return cls._read_dsv(path, f_open_options, f_read_options)
        elif suffix in MAPPING_READ_CONFIG:
            return cls._read_mapping(path, f_open_options, f_read_options, **kwargs)
        else:
            return cls._read_plain(path, f_open_options)

    @staticmethod
    def _read_dsv(path, f_open_options, f_read_options):
        import csv

        FileStream._prepare_io_options(
            [
                (f_open_options, "newline", ""),
                (f_read_options, "delimiter", DSV_CONFIG[path.suffix]["delimiter"]),
            ]
        )
        file_handler = open(path, **f_open_options)
        return file_handler, tuple(csv.DictReader(file_handler, **f_read_options))

    @staticmethod
    def _read_mapping(path, f_open_options, f_read_options, **kwargs):
        config = MAPPING_READ_CONFIG[path.suffix]
        load = getattr(importlib.import_module(config["import_mod"]), config["callable"])
        FileStream._prepare_io_options([(f_open_options, "mode", config["read_mode"])])

        file_handler = open(path, **f_open_options)
        content = load(file_handler, **f_read_options)
        if path.suffix == ".xml":
            if kwargs.get("include_root"):
                return file_handler, content
            # NB: return dict (instead of dict_view) to re-map it later as DictItem records
            return file_handler, next(iter(content.values()))
        return file_handler, content

    @staticmethod
    def _read_plain(path, f_open_options):
        file_handler = open(path, **f_open_options)
        return file_handler, (line for line in file_handler)

    # ### writing to file ###
    def save(
        self,
        file_path=None,
        *,
        f_open_options=None,
        f_write_options=None,
        null_handler=None,
        **kwargs,
    ):
        """Writes Stream to a new file (or updates an existing one) with advanced 'writing' options passed by the user"""
        path, tmp_path = self._prepare_file_paths(file_path)

        f_open_options = f_open_options or {}
        f_write_options = f_write_options or {}

        if (suffix := path.suffix) in DSV_CONFIG:
            return self._write_dsv(path, tmp_path, f_open_options, f_write_options, null_handler)
        elif suffix in MAPPING_WRITE_CONFIG:
            return self._write_mapping(
                path, tmp_path, f_open_options, f_write_options, null_handler, **kwargs
            )
        else:
            return self._write_plain(path, tmp_path, f_open_options, f_write_options)

    def _write_dsv(self, path, tmp_path, f_open_options, f_write_options, null_handler=None):
        import csv

        if null_handler:
            self.map(null_handler)
        output = self.map(lambda x: Stream(x).to_dict()).to_tuple()

        self._prepare_io_options(
            [
                (f_open_options, "mode", "w"),
                (f_write_options, "delimiter", DSV_CONFIG[path.suffix]["delimiter"]),
                (f_write_options, "fieldnames", output[0].keys() if output else ()),
            ]
        )
        with self._atomic_write(path, tmp_path, f_open_options) as f:  # noqa
            writer = csv.DictWriter(f, **f_write_options)
            writer.writeheader()
            writer.writerows(output)

    def _write_mapping(
        self, path, tmp_path, f_open_options, f_write_options, null_handler=None, **kwargs
    ):
        config = MAPPING_WRITE_CONFIG[path.suffix]
        if existing_null_handler := null_handler or config["default_null_handler"]:
            self.map(existing_null_handler)  # noqa

        output = self.to_dict()

        io_opts_setting = [(f_open_options, "mode", config["write_mode"])]
        if path.suffix == ".xml":
            root = kwargs.get("xml_root", "root")
            output = {root: output}
            io_opts_setting.append((f_write_options, "pretty", True))
        self._prepare_io_options(io_opts_setting)

        dump = getattr(importlib.import_module(config["import_mod"]), config["callable"])
        with self._atomic_write(path, tmp_path, f_open_options) as f:  # noqa
            dump(output, f, **f_write_options)

    def _write_plain(self, path, tmp_path, f_open_options, f_write_options):
        self._prepare_io_options([(f_open_options, "mode", "w")])

        output = self.to_string(f_write_options.pop("delimiter", "\n"))
        header = f_write_options.pop("header", "")
        footer = f_write_options.pop("footer", "")
        if header or footer:
            output = f"{header}{output}{footer}"

        with self._atomic_write(path, tmp_path, f_open_options) as f:  # noqa
            f.writelines(output)

    # ### helpers ###
    @staticmethod
    def _get_file_path(file_path, read_mode=True):
        path = Path(file_path)
        if read_mode and not path.exists():
            raise FileNotFoundError(f"No such file or directory: '{file_path}'")
        if path.is_dir():
            raise IsADirectoryError(f"Given path '{file_path}' is a directory")
        return path

    def _prepare_file_paths(self, file_path):
        if file_path is None:
            file_path = self._file_path
        path = self._get_file_path(file_path, read_mode=False)
        tmp_path = Path(TEMP_PATH.format(file_path=self._file_path))
        if tmp_path.exists():
            # So sorry Montessori...
            tmp_path.unlink(missing_ok=True)
        return path, tmp_path

    @staticmethod
    def _prepare_io_options(settings):
        for options, key, value in settings:
            options.setdefault(key, value)

    @contextmanager
    def _atomic_write(self, path, tmp_path, f_open_options):
        try:
            if f_open_options["mode"] == "a":
                tmp_path = shutil.copyfile(path, tmp_path)

            with open(tmp_path, **f_open_options) as f:
                yield f
            shutil.move(tmp_path, path)
        except (IOError, Exception) as e:
            tmp_path.unlink(missing_ok=True)
            raise e



from .exception import (
    IllegalStateError as IllegalStateError,
    NoneTypeError as NoneTypeError,
    NoSuchElementError as NoSuchElementError,
    UnsupportedTypeError as UnsupportedTypeError,
)



from collections.abc import Mapping
from functools import wraps

from pyrio.utils import DictItem


def map_dict_items(func):
    """Converts Mapping arguments to DictItem generators"""

    @wraps(func)
    def wrapper(*args, **kw):
        if not any(isinstance(arg, Mapping) for arg in args):
            return func(*args, **kw)

        remapped = []
        for arg in args:
            if isinstance(arg, Mapping):
                remapped.append((DictItem(k, v) for k, v in arg.items()))
            else:
                remapped.append(arg)
        return func(*remapped, **kw)

    return wrapper



import itertools as it
import operator

from pyrio.utils import Optional


class ItertoolsMixin:
    """Provides integration with itertools methods; pass corresponding parameters as kwargs"""

    NO_SIGNATURE_FUNCTIONS = ["chain", "islice", "product", "repeat", "zip_longest"]
    NO_KWARGS_FUNCTIONS = ["dropwhile", "filterfalse", "starmap", "takewhile", "tee"]

    def use(self, it_function, **kwargs):
        """Provides integration with itertools methods; pass corresponding parameters as kwargs"""
        import inspect

        if self._handle_no_signature_functions(it_function, **kwargs):
            return self

        signature = inspect.signature(it_function).parameters
        if self._handle_no_kwargs_functions(signature, it_function, **kwargs):
            return self

        return self._handle_default_signature_functions(signature, it_function, **kwargs)

    def _handle_no_signature_functions(self, it_function, **kwargs):
        if it_function.__name__ not in self.NO_SIGNATURE_FUNCTIONS:
            return False

        if it_function.__name__ in ("product", "zip_longest"):
            if isinstance(self.iterable, range):
                self.iterable = it_function(self.iterable, **kwargs)
            else:
                self.iterable = it_function(*self.iterable, **kwargs)
            return True

        # functions like 'chain' don't expect key-word arguments
        self.iterable = it_function(self.iterable, *kwargs.values())
        return True

    def _handle_no_kwargs_functions(self, signature, it_function, **kwargs):
        # handle functions that take only iterable as arg
        if len(signature.keys()) == 1 and "iterable" in signature:
            self.iterable = it_function(self.iterable)
            return True

        # handle functions that take no kwargs
        if it_function.__name__ in self.NO_KWARGS_FUNCTIONS:
            if it_function.__name__ == "tee":
                self.iterable = it_function(self.iterable, *kwargs.values())
            else:
                self.iterable = it_function(*kwargs.values(), self.iterable)
            return True
        return False

    def _handle_default_signature_functions(self, signature, it_function, **kwargs):
        if "iterable" in signature:
            kwargs["iterable"] = self.iterable
        elif "data" in signature:
            kwargs["data"] = self.iterable
        self.iterable = it_function(**kwargs)
        return self

    # ### 'recipes' ###
    # https://docs.python.org/3/library/itertools.html#itertools-recipes
    def tabulate(self, mapper, start=0):
        """Returns function(0), function(1), ..."""
        self.iterable = map(mapper, it.count(start))
        return self

    def repeat_func(self, operation, times=None):
        """Repeats calls to func with specified arguments"""
        self.iterable = it.starmap(operation, it.repeat(self.iterable, times=times))
        return self

    def ncycles(self, count=0):
        """Returns the stream elements n times"""
        self.iterable = it.chain.from_iterable(it.repeat(tuple(self.iterable), count))
        return self

    def consume(self, n=None):
        """Advances the iterator n-steps ahead. If n is None, consumes stream entirely"""
        import collections

        if n is None:
            self.iterable = collections.deque(self.iterable, maxlen=0)
            return self
        if n < 0:
            raise ValueError("Consume boundary cannot be negative")
        self.iterable = it.islice(self.iterable, n, len(self.iterable))
        return self

    def take_nth(self, idx, default=None):
        """Returns Optional with the nth element of the stream or a default value"""
        if idx < 0:
            idx = len(self.iterable) + idx
        return Optional.of_nullable(next(it.islice(self.iterable, idx, None), default))

    def all_equal(self, key=None):
        """Returns True if all elements of the stream are equal to each other"""
        return len(list(it.islice(it.groupby(self.iterable, key), 2))) <= 1

    def view(self, start=0, stop=None, step=None):
        """Provides access to a selected part of the stream"""
        if start < 0:
            start = len(self.iterable) + start

        if stop and stop < 0:
            stop = len(self.iterable) + stop

        if step and step < 0:
            raise ValueError("Step must be a positive integer or None")

        self.iterable = it.islice(self.iterable, start, stop, step)
        return self

    # ### unique ###
    def unique(self, key=None, reverse=False):
        """Yields unique elements in sorted order. Supports unhashable inputs"""
        self.iterable = self._unique(sorted(self.iterable, key=key, reverse=reverse), key=key)
        return self

    @staticmethod
    def _unique(iterable, key=None):
        return map(next, map(operator.itemgetter(1), it.groupby(iterable, key)))

    def unique_just_seen(self, key=None):
        """Yields unique elements, preserving order. Remembers only the element just seen"""
        self.iterable = map(next, map(operator.itemgetter(1), it.groupby(self.iterable, key)))
        return self

    def unique_ever_seen(self, key=None):
        """Yields unique elements, preserving order. Remembers all elements ever seen"""
        self.iterable = self._unique_ever_seen(self.iterable, key)
        return self

    @staticmethod
    def _unique_ever_seen(iterable, key=None):
        seen = set()
        for element in iterable:
            k = key(element) if key else element
            if k not in seen:
                seen.add(k)
                yield element

    # ### ###
    def sliding_window(self, n):
        """Collects data into overlapping fixed-length chunks or blocks"""
        if n < 0:
            raise ValueError("Window size cannot be negative")
        self.iterable = self._sliding_window(self.iterable, n)
        return self

    @staticmethod
    def _sliding_window(iterable, n):
        import collections

        window = collections.deque(it.islice(iterable, n - 1), maxlen=n)
        for x in it.islice(iterable, n - 1, len(iterable)):
            window.append(x)
            yield tuple(window)

    def grouper(self, n, *, incomplete="fill", fill_value=None):
        """Collects data into non-overlapping fixed-length chunks or blocks"""
        self.iterable = self._grouper(n, incomplete, fill_value)
        return self

    def _grouper(self, n, incomplete="fill", fill_value=None):
        iterators = [iter(self.iterable)] * n
        match incomplete:
            case "fill":
                return it.zip_longest(*iterators, fillvalue=fill_value)
            case "strict":
                return zip(*iterators, strict=True)
            case "ignore":
                return zip(*iterators)
            case _:
                raise ValueError(
                    f"Invalid incomplete flag '{incomplete}', expected: 'fill', 'strict', or 'ignore'"
                )

    def round_robin(self):
        """Visits input iterables in a cycle until each is exhausted"""
        self.iterable = self._round_robin(self.iterable)
        return self

    @staticmethod
    def _round_robin(iterable):
        # Algorithm credited to George Sakkis
        iterators = map(iter, iterable)
        for num_active in range(len(iterable), 0, -1):
            iterators = it.cycle(it.islice(iterators, num_active))
            yield from map(next, iterators)

    def partition(self, predicate):
        """
        Partitions entries into true and false entries.
        Returns a stream of two nested generators
        """
        true_iter, false_iter = it.tee(self.iterable)
        self.iterable = filter(predicate, true_iter), it.filterfalse(predicate, false_iter)
        return self

    def subslices(self):
        """Returns all contiguous non-empty sub-slices"""
        slices = it.starmap(slice, it.combinations(range(len(self.iterable) + 1), 2))
        self.iterable = map(operator.getitem, it.repeat(self.iterable), slices)  # noqa
        return self

    def find_indices(self, value, start=0, stop=None):
        """Returns indices where a value occurs in a sequence or iterable"""
        self.iterable = self._find_indices(self.iterable, value, start, stop)
        return self

    @staticmethod
    def _find_indices(iterable, value, start=0, stop=None):
        iterator = it.islice(iterable, start, stop)
        for i, element in enumerate(iterator, start):
            if element is value or element == value:
                yield i



from .streams.stream import Stream as Stream
from .streams.file_stream import FileStream as FileStream
from .utils.optional import Optional as Optional
from .utils.dict_item import DictItem as DictItem

__all__ = ["Stream", "FileStream", "Optional", "DictItem"]



class DictItem:
    """Helper record class for mapping key-value pairs"""

    def __init__(self, key, value):
        self._key = key
        self._value = value

    @property
    def key(self):
        return self._key

    @property
    def value(self):
        return self._map(self._value)

    def _map(self, val):
        if isinstance(val, dict):
            return tuple(DictItem(k, self._map(v)) for k, v in val.items())
        return val

    def __repr__(self):
        key = f"{self.key}" if isinstance(self.key, str) else self.key
        value = f"{self.value}" if isinstance(self.value, str) else self.value
        return f"DictItem({key=}, {value=})"

    def __eq__(self, other):
        if not isinstance(other, DictItem):
            raise TypeError(f"{other} is not a DictItem")
        return self._key == other._key and self._value == other._value  # noqa

    def __hash__(self):
        try:
            return hash((self._key, self._value))
        except TypeError as e:
            raise TypeError(
                f"unhashable type: 'DictItem' (value of type '{type(self._value).__name__}' is unhashable)"
            ) from e
